/*
* 286CORE.C                                               Version 2.31
*
* For use with Phar Lap 286 DOS Extender.
*
* !!! THIS MODULE IS NOT OPTIONAL !!!
*
* CORE MODULE. Includes smx configuration table, main entry point,
* initialization, idle task, timeout task, exit task, C versions of
* interrupt service routines.
*
* Copyright (c) 1990-1993 Micro Digital Associates, Inc. 
*               All Rights Reserved.
*
* Author: Ralph Moore
*
***********************************************************************/

#include    "core.h"
#include    "conf.h"
#include    <bios.h>
#include    <conio.h>
#include    <dos.h>
#include    <stdlib.h>
#include    <stdio.h>

#ifdef   __TURBOC__
 #define _bios_keybrd bioskey
 #undef  outp   /*(forces fct rather than macro to avoid warning)*/
 #define _outp outp
#endif


#define  DIVR(x,y) ((2*((x)%(y)) < (y)) ? (x)/(y) : ((x)/(y))+1)


/* NOTE: Some data types used in this module are defined in phapi.h. */

/* function prototypes */

#if __cplusplus
 extern "C" {
#endif

void        ainit(void);
void        do_tick(void);
void        idle_main(void);
void        idle_state(void);
void near   keep_time(void);
void        set_indos_ptr(void);
void        sp_exit(void);
void        sp_init(void);
void far    tick_int(void);
void        timeout_init(void);
void        timeout_main(void);

void interrupt far new_atdisk(XREGS r);
void interrupt far new_doserr24(XREGS r);
void interrupt far new_xtdisk(XREGS r);
void interrupt far tick(XREGS r);
void interrupt far tick_sim(XREGS r);
void interrupt far new_gpf(EXCEP_FRAME e);
void interrupt far new_sf(EXCEP_FRAME e);

#if __cplusplus
}
#endif

/* variables */

TCB_PTR           exitx;         /* end operation task */
word        near  ictr;          /* idle tick counter */
SCB_PTR     near  in_clib;       /* in C library call */ 
SCB_PTR     near  in_dos;        /* in MS-DOS call */
byte far *  near  indos_ptr;     /* pointer to INDOS flag */
struct DAR  near  ndar;          /* ndar structure */
CB_PTR      near  null_cb;       /* null control block */
word        near  octr;          /* overhead tick counter */
int         near  pcticks;       /* counts ticks for tick simulation */
word              speed_up;      /* tick speed up factor */
word        near  tick_count;    /* counts ticks for tick speed up */
word        near  tick_ctr;      /* counts ticks for ts signalling */
word        near  wctr;          /* (useful) work tick counter */


/* array pointers to aid debugging. (NOTE: array sizes must be specified
   for code debuggers.) */

#ifdef TM
struct EREC     (far *e)[EB_SIZE];        /* eb array ptr */
struct LQ_CELL (near *l)[LQ_SIZE];        /* lq array ptr */
struct XCB     (near *r)[MAX_PRIORITY+1]; /* rq array ptr */
#endif

/* old isr pointers. (Assumed to be near by assembly routines.) */

PIHANDLER near old_atdisk_prot;
REALPTR   near old_atdisk_real;
PIHANDLER near old_timer_prot;
REALPTR   near old_timer_real;
PIHANDLER near old_xtdisk_prot;
REALPTR   near old_xtdisk_real;
PIHANDLER near old_doserr_prot;
REALPTR   near old_doserr_real;
PEHANDLER near old_gpf;
PEHANDLER near old_sf;

/* address markers -- see ddseg.asm */

extern byte near  bnheap;        /* beginning of near heap */
extern byte near  bndar;         /* beginning of near dar space */
extern byte near  enheap;        /* end of near heap */
extern byte near  endar;         /* end of near dar space */

#if    LARGE_DATA
extern byte far   bstack;        /* beginning of stack pool */
extern byte far   estack;        /* end of stack pool */
#else
extern byte       bstack;        /* beginning of stack pool */
extern byte       estack;        /* end of stack pool */
#endif

extern byte       key_table[84][4]; /* keyboard conversion table */ 
extern PXCB_PTR   sp_pipe;          /* smxProbe input pipe */
extern byte       temp_heap[];

/* default configuration values -- see ddconf.h */

struct CONF near   cf =  {
         /* max_priority */   MAX_PRIORITY,
         /* lq_size      */   LQ_SIZE,
         /* sec          */   SEC,
         /* num_blocks   */   NUM_BLOCKS,
         /* num_buckets  */   NUM_BUCKETS,
         /* num_msgs     */   NUM_MSGS,
         /* num_pools    */   NUM_POOLS,
         /* num_qlevels  */   NUM_QLEVELS,
         /* num_pipes    */   NUM_PIPES,
         /* num_tasks    */   NUM_TASKS,
         /* stack_size   */   STACK_SIZE,
         /* num_stacks   */   NUM_STACKS,
         /* eb_size      */   EB_SIZE,
         /* sp_min       */   &bstack,
         /* sp_max       */   &estack,
         /* nheap_min    */   &bnheap,
         /* nheap_max    */   &enheap,
         /* ndar_min     */   &bndar, 
         /* ndar_max     */   &endar, 
         /* fheap_min    */   0,
         /* fheap_max    */   0};


/******* MAIN
* This is the entry point from the C compiler startup code.
***********************************************************************/

void _cdecl main(int argc, char *argv[])
{
   if(argc > 1)
      speed_up = atoi(argv[1]);
   else
      speed_up = 1;
   cf.sec *= speed_up;
   coldstart();
}


/******* APPLICATION INITIALIZATION
* This is the starting point for the task environment. go_smx() (called
* at the end of coldstart) assumes ainit() to exist. This function is
* executed under the idle task. Interrupts should be masked until 
* initialization is complete, since smx calls (e.g. go_smx()) set the
* interrupt enable flag, IF. At the end of ainit(), the idle task code
* is changed to idle_main() and its priority is set to MIN. 
***********************************************************************/

void _cdecl ainit(void)
{
   word  i, val;
   static char buf[10];

   /* debug pointers, use: *e to examine error buffer
                           *l to examine lsr queue
                           *r to examine ready queue */
   #ifdef   TM
    e = (struct EREC (far *)[EB_SIZE])ebi;
    l = (struct LQ_CELL (near *)[LQ_SIZE])lqi;
    r = (struct XCB (near *)[MAX_PRIORITY+1])rq;
   #endif

   dosx_init();
   sp_init();

   #if defined _MSC_VER || !defined DEBUG
    kbd_init();
   #endif

   if(xerrno) exitx_main(xerrno); /* (so error message will be seen) */

   clr_screen(LIGHTGRAY, BLACK);
   bios_cursor_off();
   set_indos_ptr();
   unlock();         /* so can use smxProbe in ainit */

   /* DON'T USE SMXPROBE BEFORE HERE. */


   /* Show first segment used by application */

   wr_string(0,13,WHITE,BLACK,!BLINK,"first selector =");
   for(i=0; i<10; i++)
      buf[i] = 0;
   itoa(first_sel, buf, 10);
   ws(17,13,buf);

   DosSetPassToProtVec(0x24, (PIHANDLER)new_doserr24, &old_doserr_prot,
                              &old_doserr_real);
   DosSetExceptionHandler(12, new_sf, &old_sf);
   DosSetExceptionHandler(13, new_gpf, &old_gpf);

   ndar.dnext = cf.ndar_min;   /* set up ndar for create_pool() calls */
   ndar.dmax  = cf.ndar_max;

   #ifdef DEBUG
    DosSetPassToProtVec(8, (PIHANDLER)tick_sim, &old_timer_prot,
                            &old_timer_real);
   #else
    DosSetPassToProtVec(8, (PIHANDLER)tick, &old_timer_prot, &old_timer_real);
   #endif

   DosSetRealVec(0xf1, old_timer_real, NULL);

   if(speed_up > 1)
      {
      _disable();
      _outp(0x43, 0x36);     /* set timer 0 to mode 3 */
      for (i=0; i<2; i++);	 /* delay */
      val = (word)(65536L/speed_up);
      _outp(0x40, (val & 0x00FF)); /* load LSB of val */
      for (i=0; i<2; i++);			 /* delay */
      _outp(0x40, (val >> 8));	    /* load MSB of val */
      _enable();
      }

   exitx = create_task((void (*)(void))exitx_main, MAX, 0);

   in_dos = create_sem(1,0);  /* to be used only in dos_trap() */
   signal(in_dos);

   /* NOTE: The following are inhibited for DEBUG because not necessary
      and cause difficulty if exit from debugger without restoring. */

   #ifndef  DEBUG
    DosSetPassToProtVec(0x76, (PIHANDLER)new_atdisk, &old_atdisk_prot,
                                                     &old_atdisk_real);
    DosSetRealVec(0xf3, old_atdisk_real, NULL);
    DosSetPassToProtVec(0x0d, (PIHANDLER)new_xtdisk, &old_xtdisk_prot,
                                                     &old_xtdisk_real);
    DosSetRealVec(0xf4, old_xtdisk_real, NULL);
   #endif

   in_clib = create_sem(1,0); /* to be used for non-reen. C lib. calls */
   signal(in_clib);

   appl_init();   

   /* Build handle table for smxProbe. (These are nop's if smxProbe is
      not present.) */

   BUILD_HT(in_dos, "in_dos");
   BUILD_HT(in_clib,"in_clib");
   BUILD_HT(exitx,  "exitx");
   BUILD_HT((VOID_PTR)keep_time, "keep_time");

   NEW_START(self, idle_main, MIN);  /* become normal idle task */
}

      
/******* IDLE
* The idle task runs when there is no other task to run. It can be used
* to compute processor utilization as well as to gather other informa-
* tion of interest. It can also be used to perform confidence tests
* such as heap checks, stack checks, etc.
***********************************************************************/

void _cdecl idle_main()
{
   static char buffer[10];
   static int  percent_idle;  /* % of time in idle task */ 
   static int  percent_ovh;   /* % of time in smx */
   static int  percent_work;  /* % of time doing useful work */
   word        total;

   busy = 1;         /* system is idle whenever this task is running */
   hook(idle_state); /* set busy flag to idle whenever resume idle */
   unlock();         /* permit this task to be preempted */

   while(TRUE)
      {
      total = octr + ictr + wctr;   /* total ticks since last profile */

      /* Calculate and output percentages. NOTE: The following could be
         done more easily using floating point numbers. However, in many
         applications, it is undesirable to introduce floating point. */

      if(total >= 30)               /* time for new profile? */
         {
         if(percent_work == 0)      /* first time */
            {
            percent_ovh  = DIVR(1000*octr,total);
            percent_idle = DIVR(1000*ictr,total);
            percent_work = DIVR(1000*wctr,total);
            }
         else                       /* subsequent times -- average */
            {
            percent_ovh  = DIVR((DIVR(1000*octr,total) + 3*percent_ovh),4);
            percent_idle = DIVR((DIVR(1000*ictr,total) + 3*percent_idle),4);
            percent_work = DIVR((DIVR(1000*wctr,total) + 3*percent_work),4);
            }

         ictr = octr = wctr = 0;    /* clear counters */

         test(in_clib, INF);
         sprintf(buffer, "%3d", percent_idle/10);
         wr_string(9,15,YELLOW,BLACK,!BLINK,buffer);
         sprintf(buffer, ".%1d", percent_idle%10);
         wr_string(12,15,YELLOW,BLACK,!BLINK,buffer);

         sprintf(buffer, "%3d", percent_work/10);
         wr_string(9,16,YELLOW,BLACK,!BLINK,buffer);
         sprintf(buffer, ".%1d", percent_work%10);
         wr_string(12,16,YELLOW,BLACK,!BLINK,buffer);

         sprintf(buffer, "%3d", percent_ovh/10);
         wr_string(9,17,YELLOW,BLACK,!BLINK,buffer);
         sprintf(buffer, ".%1d", percent_ovh%10);
         wr_string(12,17,YELLOW,BLACK,!BLINK,buffer);
         signal(in_clib);
         }

       do_tick();
      }
}


/******* TIMEOUT
* This task is activated by a signal to the ts semaphore. (See the
* keep_time lsr.) It searches for tasks which have timed out (i.e. 
* task timer > 0 and <= etime.) When found, such a task is put into
* the ready queue and its timer is cleared. This is repeated until all
* timers have been checked. go_smx() expects timeout_init() to exist.
***********************************************************************/

void _cdecl timeout_init()
{
   self->fun = conv_code_ptr(timeout_main);
   self->priority = HI;
   test_stop(ts, INF);
}

void _cdecl timeout_main()
{
   register word     tn;
   register TCB_PTR  task;

   unlock();
   
   while(test(ts, INF))
      {
      show_etime();

      for( tn = 0, task = tcbi ; task <= tcbx ; tn++, task++)
        if(timer[tn] != 0L)
           resume_to(task);

   /* If the most significant bit of etime is set, clear it and clear the
      most significant bit in all timers. NOTE: ok to use etime directly
      since only one bit is being tested and changed. */
    
      if(etime & 0x80000000L)
         {
         for( tn = 0, task = tcbi ; task <= tcbx ; tn++, task++)
           timer[tn] &= 0x7FFFFFFFL;
         etime &= 0x7FFFFFFFL;
         }
      }
}


/****** EXIT SMX
* This function may be executed either by directly calling it or by
* starting the exitx task. Its purpose is to ensure a clean exit. In
* order to avoid exiting during a DOS call, the exitx task will wait
* up to 2 seconds for DOS to finish.
***********************************************************************/   

void _cdecl exitx_main(word errno)
{
   word  i;

   if(exitx)
      test(in_dos, 2*SEC);  /* wait up to 2 seconds for DOS op to finish */

   if(speed_up > 1)  /* restore normal tick rate */
      {
      _disable();
      _outp(0x43, 0x36);    /* set timer 0 to mode 3 */
      for (i=0; i<2; i++);	/* delay */
      _outp(0x40, 0);       /* load LSB of 0 */
      for (i=0; i<2; i++); /* delay */
      _outp(0x40, 0);	      /* load MSB of 0 */
      _enable();
      }

    if(old_timer_prot)
       DosSetRealProtVec(8, old_timer_prot, old_timer_real, NULL, NULL);
    DosSetRealProtVec(0x24, old_doserr_prot, old_doserr_real, NULL, NULL);

   #ifndef DEBUG
    DosSetRealProtVec(0x76, old_atdisk_prot, old_atdisk_real, NULL, NULL);
    DosSetRealProtVec(0x0d, old_xtdisk_prot, old_xtdisk_real, NULL, NULL);
   #endif

   quit_appl();
   kbd_exit();
   sp_exit();

   if(errno == NULL_PTR_REF)
      null_cb->fl = 0;   /* clear null error before msc calls */

   #ifdef DEBUG
    wr_string(0,22,LIGHTGRAY,BLACK,!BLINK,xerrmsg[errno]);
   #endif

   cursor_off();
   bios_cursor_on();

   while(_bios_keybrd(1))  /* empty bios keyboard buffer */
      _bios_keybrd(0);

   if(errno)
      _exit(errno);
   else
      if(_doserrno)
         _exit(_doserrno+100);
      else
         _exit(0);
}


/******* INTERRUPT SERVICE ROUTINES
* These are C versions of basic interrupt service routines. See bios.asm
* for equivalent assembly versions. C_ISRS in conf.h controls which
* is used.
***********************************************************************/

#ifndef __ZTC__		/* Zortech must use assembly ISR's	*/

/* The tick_sim isr does not require ENTER_ISR & EXIT_ISR because it 
   does no invokes and it has higher priority than any interrupt using
   enter_isr and exit_isr. */

void interrupt far tick_sim(XREGS r)
{
   pcticks++;

   if(speed_up > 1)
      if (++tick_count >= speed_up)
         {
         tick_count = 0;
         DosRealIntr(0xf1, NULL, 0, 0);
         }
      else
         _outp(0x20, 0x20);  /* EOI to interrupt controller */
	else
      DosRealIntr(0xf1, NULL, 0, 0);

   r.iregs.ax = r.iregs.ax; /* to avoid warning */
}

/* examples of "encapsulated" interrupts */
                                
void interrupt far new_atdisk(XREGS r)
{
   ENTER_ISR();
   DosRealIntr(0xf3, NULL, 0, 0);
   EXIT_ISR();
}

/* NOTE: Encapsulating the following interrupt for an AT will permit
   operation with LANtastic (if it is using irq5). */

void interrupt far new_xtdisk(XREGS r)
{
   ENTER_ISR();
   DosRealIntr(0xf4, NULL, 0, 0);
   EXIT_ISR();
}

void interrupt far tick(XREGS r)
{
   ENTER_ISR();
   INVOKE(keep_time, 0);

   /* Increment profile counters. NOTE: Profiling will be more accurate
      if the following is done for an asynchronous interrupt. */

   if(busy > 1)      
      wctr++;
   else if(busy == 1)
      ictr++;
   else
      octr++;

   /* Chain to the BIOS timer isr. */

  #ifndef DEBUG
   if(speed_up > 1)
      if (++tick_count >= speed_up)
         {
         tick_count = 0;
         DosRealIntr(0xf1, NULL, 0, 0);
         }
      else
         _outp(0x20, 0x20);  /* EOI to interrupt controller */
   else
      DosRealIntr(0xf1, NULL, 0, 0);
  #endif

  #ifdef DEBUG
   EXIT_ISR_DEBUG();
  #else
   EXIT_ISR();
  #endif

   r.iregs.ax = r.iregs.ax; /* to avoid warning */
}

/******* KEEP TIME LSR
* NOTE: If using the following lsr and using Microsoft C, add /NT_TEXT
* to the compiler command line for this module. For Turbo C, add -zC_TEXT.
***********************************************************************/
       
void _cdecl near keep_time()
{
  #ifdef BIOS_KBD
   uint char_code;
   byte key;

   if(_bios_keybrd(_NKEYBRD_READY))
      {
      char_code = _bios_keybrd(_NKEYBRD_READ);

      key = (byte)char_code;
      if(key == 0)  /* if extended code, get auxiliary code: */
         {
         key = (byte)(char_code >> 8); /* get upper byte */
         key = key_table[key][0];      /* translate cursor key */
         }

       if((key == Ctrl_S) && sp_present)     /* swap */
          if(sp_scrn_sel)
             pput_char(key, op_pipe, NO_WAIT);
          else 
             pput_char(key, sp_pipe, NO_WAIT);
       else

       pput_char(key, kbd_pipe, NO_WAIT);
      }
  #endif

   #if TIME_SLICE == 1
    if(rq->fl)
       bump_task((TCB_PTR)rq->fl,0);
   #endif

   etime++;
   signal(ticks);
   if(++tick_ctr >= cf.sec)
      {
      tick_ctr = 0;
      stime++;
      signal(ts);
      }

}     

#endif /* __ZTC__ */


/******* SET INDOS POINTER
* The following function sets a pointer to the INDOS flag maintained
* by DOS. This is used to prevent DOS conflicts with tsrs.
**********************************************************************/

void  set_indos_ptr(void)
{
#ifdef _MSC_VER
 static   union  _REGS  regs;
 static   struct _SREGS sregs;
#else
 static   union  REGS  regs;
 static   struct SREGS sregs;
#endif

   regs.h.ah = 0x34;
   #ifdef _MSC_VER
    _intdosx(&regs, &regs, &sregs);
   #else
    intdosx(&regs, &regs, &sregs);
   #endif
   FP_SEG(indos_ptr) = sregs.es;
   FP_OFF(indos_ptr) = regs.x.bx;
}


/******* HEAP TRANSITION ROUTINES
* The following functions convert compiler heap calls to smx heap calls
* at link time. To enable, define USE_SMX_HEAP. See xheap.h for macros
* to convert at compile time.
***********************************************************************/
// #define USE_SMX_HEAP     

#ifdef   USE_SMX_HEAP

void *calloc(size_t nitems, size_t size)
{
   #if !LARGE_DATA
	 return (_ncallocx(nitems, size));
   #else
	 return (_fcallocx(nitems, size));
   #endif
}

void free(void *block)
{
   #if !LARGE_DATA
    _nfreex(block);
   #else
    _ffreex(block);
   #endif
}

#ifdef   _MSC_VER

void _ffree(void far *block)
{
   _ffreex(block);
}

void _nfree(void near *block)
{
   _nfreex(block);
}
#endif

/* malloc's are performed by compiler startup code for standard i/o
   environment space, and other puposes. At the time, smx heaps have
   not been initialized. To handle this, two approaches are shown:
   (1) simple allocation from a fixed-size buffer above ndar (so it
       won't be cleared.)
   (2) allocation from DOS. (commented out)
   The second is more flexible, but will not work for a small data
   model. The best solution is to modify the compiler startup code. */

void *malloc(size_t size)
{
   static char *this_pos, *next_pos;
//   static union REGS regs;

   next_pos  = (char *)&temp_heap;

   if(!smx_init)  /* if smx heaps not initialized */
      {
      this_pos = next_pos;
      next_pos += size;
      return(this_pos);
//		size >>= 4;    /* convert size to paragraphs */
//		size++;		
//		regs.x.bx = size;	   /* make DOS memory allocation call */
//		regs.h.ah = 0x48; 	
//		intdos(&regs, &regs); 
//      return(regs.x.cflag ? NULL : MK_FP(regs.x.ax, 0));
   	}

   #if !LARGE_DATA
    return(_nmallocx(size));
   #else
    return(*_fmallocx(size));
   #endif
}
#endif

/******* SMXPROBE INVOKE FUNCTION
* The following function permits invoking smxProbe from within a code
* debugger.
***********************************************************************/

#ifdef TM

#if __cplusplus
 extern "C" {
#endif

void     sp_call(void);
BOOLEAN  pr(void);

#if __cplusplus
}
#endif


BOOLEAN _cdecl pr(void)
{
   if(IS_LOCKED(ct))
      {
      UNLOCK();
      sp_call();
      }
   else
      {
      sp_call();
      UNLOCK();
      }
   return(TRUE);
}

#endif  /* of TM */


/* Simulate tick for operation with code debugger. Must be called from
   all priority 0 tasks. Also used by smxProbe. */

void _cdecl do_tick(void)
   {
   #ifdef DEBUG
    XREGS xr;

    if(pcticks)
       {
       pcticks--;  /* decrement before task switch if timeslicing */
       xr.iregs.ax = 0; /* to avoid warning */
       (*tick)(xr);
       }
   #endif
   }


/* Protected Mode Exception Handlers */

#if __cplusplus
 extern "C" {
#endif

char    *nameof(void near *);

#if __cplusplus
}
#endif


void interrupt far new_gpf(EXCEP_FRAME e)
{
   e = e;
   wr_string(0,3,WHITE,BLACK,!BLINK,"general protection fault");
   ws(0,4,"ct = ");
   ws(5,4,nameof(ct));
   exitx_main(13);
}


void interrupt far new_sf(EXCEP_FRAME e)
{
   e = e;
   wr_string(0,3,WHITE,BLACK,!BLINK,"stack fault");
   ws(0,4,"ct = ");
   ws(5,4,nameof(ct));
   exitx_main(12);
}

